#include "traverser.h"

#include<vector>

#include "path_checker.h"
#include "../tools/rose_nodes.h"
#include "../tools/class_is_derived_from.h"
#include "../tools/search_node_in_namespace.h"
#include "../internal_representation/design.h"

std::vector<SgClassDefinition*> risc::traverser::find_all_modules()
{
  std::vector<SgClassDefinition*> modules;

  const std::vector<SgNode*> &class_defs = tools::RoseNodes::get_class_defs();

  for(std::vector<SgNode*>::const_iterator
      iter  = class_defs.begin();
      iter != class_defs.end();
      iter++) {

    SgClassDefinition *cd = isSgClassDefinition(*iter);

    // We have to check if the class is derived from sc_module
    // If yes, we found a module in the user design space
    if(risc::tools::is_derived_from_sc_module(cd)) {

      modules.push_back(cd);
    }
  }

  return modules;
}

std::vector<SgClassDefinition*> risc::traverser::find_all_channels()
{
  std::vector<SgClassDefinition*> channels;

  const std::vector<SgNode*> &class_defs = tools::RoseNodes::get_class_defs();

  for(std::vector<SgNode*>::const_iterator
      iter  = class_defs.begin();
      iter != class_defs.end();
      iter++) {

    SgClassDefinition *cd = isSgClassDefinition(*iter);

    // We have to check if the class is derived from sc_module
    // If yes, we found a module in the user design space
    if(risc::tools::is_derived_from_sc_channel(cd) ||
        risc::tools::is_derived_from_sc_prim_channel(cd)) {

      channels.push_back(cd);
    }
  }

  return channels;
}

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
