%{

/*
 * Parser.y file
 * To generate the parser run: "bison Parser.y"
 */

%}

%code requires {

#ifndef DYNAMIC_ANALYSIS_PARSER_TYPEDEF_SCANNER_T
#define DYNAMIC_ANALYSIS_PARSER_TYPEDEF_SCANNER_T

#include <stdio.h>
#include <iostream>
#include <list>
#include <cstdlib>
#include <utility>
#include "dynamic_internal_representation.h"

extern DynamicModule *ast_root;
extern std::list<std::pair<std::string, int*> > global_variable_addresses;

void yyerror(const char *s);
int yylex();

#endif
}

%union {
  char *Address;
  char *Pointer;
  char *Name;
  DynamicObject *Element;
  DynamicModule *ModuleInstance;
  std::list<DynamicObject*> *Elements;
}

/* declaration of tokens */
%token           T_LPAREN
%token           T_RPAREN
%token           T_COLON
%token           T_DOT
%token           T_COMMA
%token           T_AT
%token           T_PERCENT
%token           T_AND
%token           T_HASH
%token           T_MINUS
%token           T_DOLLAR
%token           T_PLUS
%token           T_LESS
%token           T_GREATER
%token           T_CURLY_OPEN
%token           T_CURLY_CLOSE
%token           T_QUESTION_MARK

%token <Name>    T_RISC_NAME
%token <Pointer> T_RISC_POINTER
%token <Address> T_ADDRESS


%type <Elements> elements
%type <Element>  module



%error-verbose

%%

module:   T_RISC_NAME T_HASH T_ADDRESS T_LPAREN elements T_RPAREN T_COMMA global
          {
            /* module */

            DynamicModule *m = new DynamicModule();
            m->name_ = $1;
            long int address_as_value = strtol($3, NULL, 16);
            m->address_ = reinterpret_cast<int*>(address_as_value);

            for(std::list<DynamicObject*>::iterator
                iter  = $5->begin();
                iter != $5->end();
                iter++) {

              // Identifying what elements is
              DynamicPort *port = dynamic_cast<DynamicPort*>(*iter);
              DynamicPointer *pointer = dynamic_cast<DynamicPointer*>(*iter);
              DynamicReference *reference = dynamic_cast<DynamicReference*>(*iter);
              DynamicPrimChannel *channel = dynamic_cast<DynamicPrimChannel*>(*iter);
              DynamicVariable *variable = dynamic_cast<DynamicVariable*>(*iter);
              DynamicEvent *event = dynamic_cast<DynamicEvent*>(*iter);
              DynamicEventAndListElement *event_and_list_element = dynamic_cast<DynamicEventAndListElement*>(*iter);
              DynamicEventOrListElement *event_or_list_element = dynamic_cast<DynamicEventOrListElement*>(*iter);

              DynamicModule *module = dynamic_cast<DynamicModule*>(*iter);
              DynamicHierarchicalChannel *hierarchical_channel = dynamic_cast<DynamicHierarchicalChannel*>(*iter);

              if(hierarchical_channel) {
                module = NULL;
              }

              if(port) {
                m->ports_.push_back(port);
              } else if(pointer) {
                m->pointers_.push_back(pointer);

              } else if(reference){
                m->references_.push_back(reference);

              } else if(variable){
                m->variables_.push_back(variable);

              } else if(module){
                m->sub_modules_.push_front(module);

              } else if(channel){
                m->sub_channels_.push_back(channel);

              } else if(hierarchical_channel){
                m->sub_hierarchical_channels_.push_back(hierarchical_channel);

              } else if(event){
                m->events_.push_back(event);

              } else if(event_and_list_element) {

                bool found = false;
                for(std::list<DynamicEventAndList*>::iterator
                     iter  = m->event_and_lists_.begin();
                     iter != m->event_and_lists_.end();
                     iter++) {

                  if((*iter)->address_ == event_and_list_element->address_) {
                    found = true;

                    (*iter)->events_.push_back(event_and_list_element->event_);
                  }
                }
                if(found == false) {
                  DynamicEventAndList *new_list = new DynamicEventAndList();
                  new_list->name_ = event_and_list_element->name_;
                  new_list->address_ = event_and_list_element->address_;
                  new_list->events_.push_back(event_and_list_element->event_);

                  m->event_and_lists_.push_back(new_list);
                }
              } else if(event_or_list_element) {

                bool found = false;
                for(std::list<DynamicEventOrList*>::iterator
                     iter  = m->event_or_lists_.begin();
                     iter != m->event_or_lists_.end();
                     iter++) {

                  if((*iter)->address_ == event_or_list_element->address_) {
                    found = true;

                    (*iter)->events_.push_back(event_or_list_element->event_);
                  }
                }
                if(found == false) {
                  DynamicEventOrList *new_list = new DynamicEventOrList();
                  new_list->name_ = event_or_list_element->name_;
                  new_list->address_ = event_or_list_element->address_;
                  new_list->events_.push_back(event_or_list_element->event_);

                  m->event_or_lists_.push_back(new_list);
                }
              } else {
                // empty rule
              }
            }

            ast_root = m;

            $$ = m;
          }

        | T_RISC_NAME T_QUESTION_MARK T_ADDRESS T_LPAREN elements T_RPAREN T_COMMA global
          {
            /* hierarchical channel */

            DynamicHierarchicalChannel *hc = new DynamicHierarchicalChannel();
            hc->name_ = $1;
            long int address_as_value = strtol($3, NULL, 16);
            hc->address_ = reinterpret_cast<int*>(address_as_value);

            for(std::list<DynamicObject*>::iterator
                iter  = $5->begin();
                iter != $5->end();
                iter++) {

              // Identifying what elements is
              DynamicPort *port = dynamic_cast<DynamicPort*>(*iter);
              DynamicPointer *pointer = dynamic_cast<DynamicPointer*>(*iter);
              DynamicReference *reference = dynamic_cast<DynamicReference*>(*iter);
              DynamicModule *module = dynamic_cast<DynamicModule*>(*iter);
              DynamicPrimChannel *channel = dynamic_cast<DynamicPrimChannel*>(*iter);
              DynamicHierarchicalChannel *hierarchical_channel = dynamic_cast<DynamicHierarchicalChannel*>(*iter);
              DynamicVariable *variable = dynamic_cast<DynamicVariable*>(*iter);
              DynamicEvent *event = dynamic_cast<DynamicEvent*>(*iter);
              DynamicEventAndListElement *event_and_list_element = dynamic_cast<DynamicEventAndListElement*>(*iter);
              DynamicEventOrListElement *event_or_list_element = dynamic_cast<DynamicEventOrListElement*>(*iter);

              if(port) {
                hc->ports_.push_back(port);
              } else if(pointer) {
                hc->pointers_.push_back(pointer);

              } else if(reference){
                hc->references_.push_back(reference);

              } else if(variable){
                hc->variables_.push_back(variable);

              } else if(module){
                hc->sub_modules_.push_front(module);

              } else if(channel){
                hc->sub_channels_.push_back(channel);

              } else if(hierarchical_channel){
                hc->sub_hierarchical_channels_.push_back(hierarchical_channel);

              } else if(event){
                hc->events_.push_back(event);

              } else if(event_and_list_element) {

                bool found = false;
                for(std::list<DynamicEventAndList*>::iterator
                     iter  = hc->event_and_lists_.begin();
                     iter != hc->event_and_lists_.end();
                     iter++) {

                  if((*iter)->address_ == event_and_list_element->address_) {
                    found = true;

                    (*iter)->events_.push_back(event_and_list_element->event_);
                  }
                }
                if(found == false) {
                  DynamicEventAndList *new_list = new DynamicEventAndList();
                  new_list->name_ = event_and_list_element->name_;
                  new_list->address_ = event_and_list_element->address_;
                  new_list->events_.push_back(event_and_list_element->event_);

                  hc->event_and_lists_.push_back(new_list);
                }
              } else if(event_or_list_element) {

                bool found = false;
                for(std::list<DynamicEventOrList*>::iterator
                     iter  = hc->event_or_lists_.begin();
                     iter != hc->event_or_lists_.end();
                     iter++) {

                  if((*iter)->address_ == event_or_list_element->address_) {
                    found = true;

                    (*iter)->events_.push_back(event_or_list_element->event_);
                  }
                }
                if(found == false) {
                  DynamicEventOrList *new_list = new DynamicEventOrList();
                  new_list->name_ = event_or_list_element->name_;
                  new_list->address_ = event_or_list_element->address_;
                  new_list->events_.push_back(event_or_list_element->event_);

                  hc->event_or_lists_.push_back(new_list);
                }
              } else {
                // empty rule
              }
            }

            ast_root = hc;

            $$ = hc;
          }
      ;

elements: T_RISC_NAME T_COLON T_ADDRESS T_COMMA elements
          {
            /* port */
            DynamicPort *p = new DynamicPort();
            p->name_ = $1;

            long int address_as_value = strtol($3, NULL, 16);
            p->address_ = reinterpret_cast<int*>(address_as_value);

            $5->push_back(p);
            $$ = $5;
          }
        | T_RISC_NAME T_PLUS T_ADDRESS T_COMMA elements
          {
            /* event */
            DynamicEvent *e = new DynamicEvent();
            e->name_ = $1;

            long int address_as_value = strtol($3, NULL, 16);
            e->address_ = reinterpret_cast<int*>(address_as_value);

            $5->push_back(e);
            $$ = $5;
          }
        | T_RISC_NAME T_LESS T_ADDRESS T_CURLY_OPEN T_ADDRESS T_COMMA elements
          {
            /* event and list */
            DynamicEventAndListElement *list = new DynamicEventAndListElement();
            list->name_ = $1;

            long int address_event_list = strtol($3, NULL, 16);
            list->address_ = reinterpret_cast<int*>(address_event_list);

            long int address_event = strtol($5, NULL, 16);
            list->event_ = reinterpret_cast<int*>(address_event);

            $7->push_back(list);
            $$ = $7;
          }
        | T_RISC_NAME T_GREATER T_ADDRESS T_CURLY_CLOSE T_ADDRESS T_COMMA elements
          {
            /* event or list */
            DynamicEventOrListElement *list = new DynamicEventOrListElement();
            list->name_ = $1;

            long int address_event_list = strtol($3, NULL, 16);
            list->address_ = reinterpret_cast<int*>(address_event_list);

            long int address_event = strtol($5, NULL, 16);
            list->event_ = reinterpret_cast<int*>(address_event);

            $7->push_back(list);
            $$ = $7;
          }
        | T_RISC_NAME T_AT T_ADDRESS T_COMMA elements
          {
            /* reference */
            DynamicReference *r = new DynamicReference();
            r->name_ = $1;

            long int address_as_value = strtol($3, NULL, 16);
            r->address_ = reinterpret_cast<int*>(address_as_value);

            $5->push_back(r);
            $$ = $5;
          }
        | T_RISC_NAME T_DOLLAR T_ADDRESS T_COMMA elements
          {
            /* variable */
            DynamicVariable *r = new DynamicVariable();
            r->name_ = $1;

            long int address_as_value = strtol($3, NULL, 16);
            r->address_ = reinterpret_cast<int*>(address_as_value);

            $5->push_back(r);
            $$ = $5;
          }
        | T_RISC_NAME T_RISC_POINTER T_ADDRESS T_COMMA elements
          {
            /* pointer */
            DynamicPointer *r = new DynamicPointer();
            r->name_ = $1;

            long int address_as_value = strtol($3, NULL, 16);
            r->address_ = reinterpret_cast<int*>(address_as_value);

            $5->push_back(r);
            $$ = $5;
          }
        | T_RISC_NAME T_PERCENT T_ADDRESS T_AND T_RISC_NAME T_COMMA elements
          {
            /* channel */
            DynamicPrimChannel *c = new DynamicPrimChannel();
            c->name_ = $1;

            long int address_as_value = strtol($3, NULL, 16);
            c->address_ = reinterpret_cast<int*>(address_as_value);

            c->class_name_ = $5;

            $7->push_back(c);
            $$ = $7;
          }
        | module elements
          {
            $2->push_back($1);
            $$ = $2;
          }
        |
          {
            std::list<DynamicObject*> *result_list
              = new std::list<DynamicObject*>();
            $$ = result_list;
            /* empty rule */
          }
        ;

global: T_MINUS T_RISC_NAME T_MINUS T_ADDRESS T_COMMA global
        {
          long int address_as_value = strtol($4, NULL, 16);
          std::string name($2);

          global_variable_addresses.
            push_back(
              std::pair<std::string, int*> (
                name,
                reinterpret_cast<int*>(address_as_value)
              )
            );

        }
      |
        {
        }
      ;

%%

std::list<std::pair<std::string, int*> > global_variable_addresses;
extern FILE *yyin;
DynamicModule *ast_root;

