#ifndef MODULE_INSTANCE_H_INCLUDED_
#define MODULE_INSTANCE_H_INCLUDED_

#include <string>
#include <vector>

#include "rose.h"

#include "instance.h"

#include "class.h" // included for get_definition
#include "module.h" // included for get_definition

namespace risc {

/** \class ModuleInstance
 *  \brief This class represents a instance of a module
 *
 *  A module instance has a type pointer and rose ast pointer.
 */
class ModuleInstance: public Instance {

public:

  /**
   * \brief default constructor
   */
  explicit ModuleInstance(SgVariableDefinition *ast_node, Module *ir_def);

  /**
   * \brief This constructor is used for classes which derived from sc_module
   *        Currently, this is only sc_channel
   */
  explicit ModuleInstance(SgVariableDefinition *ast_node, Type type);


  /**
   * \fn       get_definition
   * \brief    returns this associated definition of the module instance
   */
  virtual Module* get_definition();

private:

  /**
   * \brief This pointer points to the corresponding definition in the
   *        internal representation
   */
  Module* module_definition_;

  /**
   * \brief Disabled copy constructor
   */
  ModuleInstance(const ModuleInstance &md);
};

typedef std::vector<ModuleInstance*>
            ModuleInstanceVector;
typedef std::vector<ModuleInstance*>::iterator
            ModuleInstanceVectorIter;
typedef std::vector<ModuleInstance*>::const_iterator
            ModuleInstanceVectorConstIter;

}; // end namespace risc

#endif /* MODULE_INSTANCE_H_INCLUDED_ */

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
