#ifndef CACHED_FUNCTION_AST_ATTRIBUTES_H_INCLUDED_
#define CACHED_FUNCTION_AST_ATTRIBUTES_H_INCLUDED_

#include "rose.h"

#include "segment_graph.h"

namespace risc {

namespace sg {

/**
 * \brief   This class stores precached information of a analyzed function.
 *          Each function definition has an associated precached ast attribute.
 *          So, we store the beginning vertex and the leaving segments of a
 *          vertex.
 */
class CachedFunctionAstAttributes: public AstAttribute {

public:

  enum FunctionIntegration {Inlining, Connecting};

  // TODO for the new rose release
  //virtual AstAttribute::OwnershipPolicy getOwnershipPolicy() const ROSE_OVERRIDE
  //{
  //    return CONTAINER_OWNERSHIP;
  //}

  CachedFunctionAstAttributes(VertexDescriptor starting_vertex,
      FunctionIntegration function_integration);

  /**
   * \fn       has_boundary
   * \brief    This function checks if a potential function call faces any
   *           boundary statement. If the 'leaving_segments_' set contains only
   *           the 'starting_vertex_' than there is no boundary statement.
   */
  bool has_boundary();

  /**
   * \fn       starting_segment_is_leaving_segment
   * \brief    The function checks if the starting segment is a leaving segment
   *           at the same time.
   */
  bool starting_segment_is_leaving_segment();


  FunctionIntegration function_integration_;

  VertexDescriptor starting_vertex_;
  SegmentGraph::SegmentSet leaving_segments_;
};

} // end of namespace sg

} // end of namespace risc

#endif /* CACHED_FUNCTION_AST_ATTRIBUTES_H_INCLUDED_ */

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
