#include "mapped_variable.h"

risc::sg::MappedVariable::MappedVariable(
    SgVariableSymbol* symbol, int instance_id):
  symbol_(symbol),
  instance_id_(instance_id)
{ }

risc::sg::MappedVariable::MappedVariable():
  symbol_(0),
  instance_id_(MappedVariable::InvalidVariable)
{ }

risc::sg::Conflict::Conflict(SgNode *node):
  node_(node)
{ }

risc::sg::Conflict::Conflict(MappedVariable variable):
  node_(NULL),
  variable_(variable)
{ }


int const risc::sg::MappedVariable::InvalidVariable = -2;

bool
risc::sg::MappedVariable::is_valid()
{
  return instance_id_ != MappedVariable::InvalidVariable;
}

bool
risc::sg::MappedVariable::operator==(const MappedVariable &ohs)
{
  return (symbol_ == ohs.symbol_) && (instance_id_ == ohs.instance_id_);
}

bool
risc::sg::MappedVariable::operator!=(const MappedVariable &ohs)
{
  return (symbol_ != ohs.symbol_) || (instance_id_ != ohs.instance_id_);
}

const bool
risc::sg::operator< (const MappedVariable &lhs, const MappedVariable &rhs)
{
  return lhs.symbol_ < rhs.symbol_;
}

const bool
risc::sg::operator<(const Conflict &lhs, const Conflict &rhs)
{
  // the value -2 means that variable is not set
  if((lhs.variable_.instance_id_ != MappedVariable::InvalidVariable)
      && (rhs.variable_.instance_id_ != MappedVariable::InvalidVariable)) {
    return lhs.variable_.symbol_ < rhs.variable_.symbol_;
  }

  if((lhs.node_ != NULL) && (rhs.node_ != NULL)) {
    return lhs.node_ < rhs.node_;
  }

  if((lhs.variable_.instance_id_ == MappedVariable::InvalidVariable)
      && (rhs.node_ != NULL)) {

    return lhs.variable_.symbol_ < rhs.node_;
  }

  if((lhs.node_ != NULL) &&
      (rhs.variable_.instance_id_ == MappedVariable::InvalidVariable)) {

    return lhs.node_ < rhs.variable_.symbol_;
  }

  assert(0 && "Uncovered case with conflict variable");
}

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
