#ifndef MAPPED_VARIABLE_H_INCLUDED_
#define MAPPED_VARIABLE_H_INCLUDED_

#include "rose.h"

#include<list>

namespace risc {

namespace sg {

/**
 * \struct MappedVariable
 * \brief  This is a helper class to describe a variable in the design
 *         through the symbol and the instance id
 *
 *         instance_id values
 *          0 ... n instance id
 *         -1       global variable
 *         -2       invalid variable
 */
struct MappedVariable {

  MappedVariable(SgVariableSymbol* symbol, int instance_id);

  MappedVariable();

  bool is_valid();

  bool operator==(const MappedVariable &ohs);

  bool operator!=(const MappedVariable &ohs);

  SgVariableSymbol* symbol_;
  int instance_id_;

  static int const InvalidVariable;
};

typedef std::list<MappedVariable> MappedVariableList;

const bool operator<(const MappedVariable &lhs, const MappedVariable &rhs);

struct Conflict {

  SgNode *node_;
  MappedVariable variable_;

  Conflict(SgNode *node);

  Conflict(MappedVariable variable);
};

const bool operator<(const Conflict &lhs, const Conflict &rhs);

}; // end of namesapce sg

}; // end of namespace risc

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */

#endif /* MAPPED_VARIABLE_H_INCLUDED_ */
