#include "function_helper.h"

#include "class_is_derived_from.h"
#include "file_classifier.h"
#include "../tools/rose_nodes.h"
#include "../tools/systemc_definitions.h"
#include "../segment_graph/function_annotation_attributes.h"
#include "../internal_representation/design.h"
#include "../internal_representation/dynamic_internal_representation.h"
#include "../internal_representation/hierarchical_channel.h"
#include "../internal_representation/interface.h"
#include "../internal_representation/primitive_channel.h"

extern DynamicModule *ast_root;

// TODO This one needs a major revision!
SgFunctionDeclaration*
risc::tools:: get_function_declaration_of_channel_call_statically(
    SgFunctionCallExp *func_call)
{
  SgMemberFunctionDeclaration *virtual_func_decl = NULL;


  std::vector<SgNode*> var_refs
    = NodeQuery::querySubTree(func_call->get_function(), V_SgVarRefExp);

  SgVarRefExp *port = NULL;
  for(std::vector<SgNode*>::iterator
      iter  = var_refs.begin();
      iter != var_refs.end();
      iter++) {

    SgVarRefExp *var_ref = isSgVarRefExp(*iter);

    if(tools::is_derived_from_sc_port_base(var_ref)) {

      port = var_ref;
      break;
    }
  }

  // TODO this is not clean
  SgBinaryOp *binary = isSgBinaryOp(func_call->get_function());

  SgMemberFunctionSymbol *virtual_func_symbol
    = isSgMemberFunctionRefExp(binary->get_rhs_operand())->get_symbol();
  virtual_func_decl = virtual_func_symbol->get_declaration();

#if 0
  // here we are getting the port variable.
  if(isSgArrowExp(func_call->get_function()) == NULL) {

    // in this case we are expecting a sc_fifo_in/sc_fifo_out port
    SgDotExp *dot_exp = isSgDotExp(func_call->get_function());
    SgArrowExp *arrow_exp = NULL;

    if(dot_exp) {

      arrow_exp = isSgArrowExp(dot_exp->get_lhs_operand());
      if(arrow_exp) {

        SgVarRefExp *var_ref = isSgVarRefExp(arrow_exp->get_rhs_operand());
        if(var_ref) {
          port = var_ref;
        } else {

          return NULL;
        }
      }
    }

    SgMemberFunctionSymbol *virtual_func_symbol
      = isSgMemberFunctionRefExp(dot_exp->get_rhs_operand())->get_symbol();
    virtual_func_decl = virtual_func_symbol->get_declaration();

  } else {

    // in this case we are expecting a sc_port

    // The function must be in an arrow operator like some_var->foo();
    SgArrowExp *arrow_exp = isSgArrowExp(func_call->get_function());
    SgDotExp *dot_exp = isSgDotExp(func_call->get_function());

    if(!arrow_exp) {
      if(dot_exp) {

        SgArrowExp *temp = isSgArrowExp(dot_exp->get_lhs_operand());
        if(temp) {
          arrow_exp = temp;

        } else {
          return NULL;
        }
      }
      return NULL;
    }

    // getting the port variable
    // I put this into an extra scope that the function scope is not messed up
    SgDotExp *dot_exp_2
      = isSgDotExp(
            isSgFunctionCallExp(
                arrow_exp->get_lhs_operand())->get_function());

    SgArrowExp *arrow2
      = isSgArrowExp(
            isSgCastExp(dot_exp_2->get_lhs_operand())->get_operand());

    port = isSgVarRefExp(arrow2->get_rhs_operand());

    SgMemberFunctionSymbol *virtual_func_symbol
      = isSgMemberFunctionRefExp(arrow_exp->get_rhs_operand())->get_symbol();
    virtual_func_decl = virtual_func_symbol->get_declaration();
  }
#endif


  // next we try to get the base class where the virtual function is defined
  SgClassDeclaration *channel_class_decl = NULL;
  Design *design = risc::get_design();

  if(port) {

    if(tools::is_sc_fifo_in(port) || tools::is_sc_fifo_out(port)) {

      SgTemplateInstantiationDecl *port_class_decl
        = isSgTemplateInstantiationDecl(channel_class_decl);

      const std::vector<SgNode*> &template_class_instances_
        = tools::RoseNodes::get_template_instance_decl();

      for(std::vector<SgNode*>::const_iterator
          iter  = template_class_instances_.begin();
          iter != template_class_instances_.end();
          iter++) {

        if(SageInterface::get_name(*iter)
            == SageInterface::get_name(
                   tools::SystemCDefinitions::get_sc_fifo())) {

          SgTemplateInstantiationDecl *template_class_instance
            = isSgTemplateInstantiationDecl(*iter);

          template_class_instance
            = isSgTemplateInstantiationDecl(
                  template_class_instance->get_definingDeclaration());

          SgTemplateClassDeclaration *template_class_decl
            = template_class_instance->get_templateDeclaration();

          template_class_decl
            = isSgTemplateClassDeclaration(
                  template_class_decl->get_definingDeclaration());

          if(template_class_decl->get_definition()
              == tools::SystemCDefinitions::get_sc_fifo()) {

            channel_class_decl = template_class_instance;
          }
        }
      }

    } else {

      SgType *type = port->get_symbol()->get_type();
      SgClassType *class_type = isSgClassType(type);

      if(class_type) {

        SgTemplateInstantiationDecl *temp_inst_decl
          = isSgTemplateInstantiationDecl(class_type->get_declaration());

        if(tools::is_derived_from_sc_port_base(port)) {

          SgClassType *template_type
            = isSgClassType(
                  temp_inst_decl->get_templateArguments().at(0)->get_type());

          if(template_type) {

            channel_class_decl
              = isSgClassDeclaration(template_type->get_declaration());
            channel_class_decl
              = isSgClassDeclaration(
                    channel_class_decl->get_definingDeclaration());
          }
        }
      }
    }
  }

  // A lamda function would be nice for this one
  bool has_interface = false;
  for(InterfaceVector::iterator
      iter  = design->interfaces_.begin();
      iter != design->interfaces_.end();
      iter++) {

    if((*iter)->get_ast_node() == channel_class_decl->get_definition()) {
      has_interface = true;

    }
  }

  if(has_interface) {
    for(risc::HierarchicalChannelVector::iterator
        channel_iter  = design->hierarchical_channel_.begin();
        channel_iter != design->hierarchical_channel_.end();
        channel_iter++) {

      if(tools::is_derived_from((*channel_iter)->get_ast_node(),
            channel_class_decl->get_definition())) {
        channel_class_decl = (*channel_iter)->get_ast_node()->get_declaration();
      }
    }

    for(risc::PrimitiveChannelVector::iterator
        channel_iter  = design->primitive_channel_.begin();
        channel_iter != design->primitive_channel_.end();
        channel_iter++) {
      if(tools::is_derived_from((*channel_iter)->get_ast_node(),
            channel_class_decl->get_definition())) {
        channel_class_decl = (*channel_iter)->get_ast_node()->get_declaration();
      }
    }
  }

  for(risc::HierarchicalChannelVector::iterator
      channel_iter  = design->hierarchical_channel_.begin();
      channel_iter != design->hierarchical_channel_.end();
      channel_iter++) {

    if(channel_class_decl->get_firstNondefiningDeclaration() !=
        (*channel_iter)->get_ast_node()->get_declaration()
           ->get_firstNondefiningDeclaration()) {
      continue;
    }

    // Now, we have the base class and the port variable.
    // So, we can look into the regular class and search for the same function
    // in the base and derived class
    std::vector<SgNode*> func_decls_channel
      = NodeQuery::querySubTree((*channel_iter)->get_ast_node(),
            V_SgFunctionDeclaration);

    for(std::vector<SgNode*>::iterator
        channel_func_iter  = func_decls_channel.begin();
        channel_func_iter != func_decls_channel.end();
        channel_func_iter++) {

      SgFunctionDeclaration *channel_func_decl
        = isSgFunctionDeclaration(*channel_func_iter);

      // we can only check some certain properties
      // probably, I student can write an optimized strategy for this one
      // - function name
      // - number of parameters
      // - is const function
      //
      // we cannot check ofr
      // - return type
      // - function paramter types

      bool same = true;

      if(channel_func_decl->get_name() != virtual_func_decl->get_name()) {

        same = false;
      }

      if(isSgMemberFunctionType(channel_func_decl->get_type())->isConstFunc()
          != isSgMemberFunctionType(
                 virtual_func_decl->get_type())->isConstFunc()) {

        same = false;
      }

      SgInitializedNamePtrList &parameters_func_1
        = channel_func_decl->get_parameterList()->get_args();

      SgInitializedNamePtrList &parameters_func_2
        = virtual_func_decl->get_parameterList()->get_args();

      if(parameters_func_1.size() != parameters_func_2.size()) {

        same = false;
      }

      if(channel_func_decl->get_orig_return_type()
          != virtual_func_decl->get_orig_return_type()) {

        // TODO This one needs more intensive tests
        //      This should be later covered through test cases of
        //      test_channel_analysis

        //same = false;
      }

      if(same) {
        return channel_func_decl;
      }
    }
  }


  for(risc::PrimitiveChannelVector::iterator
      channel_iter  = design->primitive_channel_.begin();
      channel_iter != design->primitive_channel_.end();
      channel_iter++) {

    if(channel_class_decl->get_firstNondefiningDeclaration() !=
        (*channel_iter)->get_ast_node()->get_declaration()
          ->get_firstNondefiningDeclaration()) {
      continue;
    }

    // Now, we have the base class and the port variable.
    // So, we can look into the regular class and search for the same function
    // in the base and derived class
    std::vector<SgNode*> func_decls_channel
      = NodeQuery::querySubTree((*channel_iter)->get_ast_node(),
            V_SgFunctionDeclaration);

    for(std::vector<SgNode*>::iterator
        channel_func_iter  = func_decls_channel.begin();
        channel_func_iter != func_decls_channel.end();
        channel_func_iter++) {


      SgFunctionDeclaration *channel_func_decl
        = isSgFunctionDeclaration(*channel_func_iter);

      // we can only check some certain properties
      // probably, I student can write an optimized strategy for this one
      // - function name
      // - number of parameters
      // - is const function
      //
      // we cannot check ofr
      // - return type
      // - function paramter types

      bool same = true;

      if(channel_func_decl->get_name() != virtual_func_decl->get_name()) {

        same = false;
      }

      if(isSgMemberFunctionType(channel_func_decl->get_type())->isConstFunc()
          != isSgMemberFunctionType(
                 virtual_func_decl->get_type())->isConstFunc()) {

        same = false;
      }

      SgInitializedNamePtrList &parameters_func_1
        = channel_func_decl->get_parameterList()->get_args();

      SgInitializedNamePtrList &parameters_func_2
        = virtual_func_decl->get_parameterList()->get_args();

      if(parameters_func_1.size() != parameters_func_2.size()) {

        same = false;
      }

      if(channel_func_decl->get_orig_return_type()
          != virtual_func_decl->get_orig_return_type()) {

        // TODO This one needs more intensive tests
        //      This should be later covered through test cases of
        //      test_channel_analysis

        //same = false;
      }

      if(same) {
        return channel_func_decl;
      }
    }
  }

  return NULL;
}

bool
risc::tools::function_call_belongs_to_a_port(SgFunctionCallExp *func_call_exp)
{
  if(func_call_exp == NULL) {
    return false;
  }

  // here we are checking if it is a sc_fifo_in or sc_fifo_out port
  SgDotExp *dot_exp = isSgDotExp(func_call_exp->get_function());
  SgArrowExp *arrow_exp = NULL;

  if(dot_exp) {

    arrow_exp = isSgArrowExp(dot_exp->get_lhs_operand());
    if(arrow_exp) {

      SgVarRefExp *var_ref = isSgVarRefExp(arrow_exp->get_rhs_operand());
      if(var_ref) {
        if(tools::is_sc_fifo_in(var_ref) || tools::is_sc_fifo_out(var_ref)) {
          return true;
        }
      }
    }
  }

  // The function must be in an arrow operator like some_var->foo();
  arrow_exp = isSgArrowExp(func_call_exp->get_function());
  dot_exp = isSgDotExp(func_call_exp->get_function());

  if(!arrow_exp) {

    if(dot_exp) {

      SgArrowExp *temp = isSgArrowExp(dot_exp->get_lhs_operand());
      if(temp) {
        arrow_exp = temp;

      } else {
        return false;
      }
    }

    return false;
  }

  SgVarRefExp *var_ref = isSgVarRefExp(arrow_exp->get_lhs_operand());
  if(var_ref) {

    SgClassType *class_type = isSgClassType(var_ref->get_type());

    SgClassDeclaration *class_decl
      = isSgClassDeclaration(class_type->get_declaration());

    class_decl = isSgClassDeclaration(class_decl->get_definingDeclaration());

    if(tools::is_derived_from_sc_port_base(class_decl->get_definition()) ) {
      return true;
    } else {
      return false;
    }
  }

  // here we are checking for a custom designed channel like port->write
  // This is in a new scope because I don't want to mess up the function scope
  {
    SgFunctionCallExp *inner_func_call_exp
      = isSgFunctionCallExp(arrow_exp->get_lhs_operand());

    if(inner_func_call_exp) {

      SgDotExp *dot_exp = isSgDotExp(inner_func_call_exp->get_function());

      if(dot_exp) {

        SgCastExp *cast_exp = isSgCastExp(dot_exp->get_lhs_operand());

        if(cast_exp) {

          SgArrowExp *arrow_exp = isSgArrowExp(cast_exp->get_operand());

          if(arrow_exp) {
            SgVarRefExp *var_ref = isSgVarRefExp(arrow_exp->get_rhs_operand());

            if(var_ref) {

              if(tools::is_derived_from_sc_port_base(var_ref)) {
                return true;
              }
            }
          }
        }
      }
    }
  }

  return false;
}

bool
risc::tools::function_call_is_in_user_domain(SgFunctionCallExp *func_call_exp)
{
  SgFunctionDeclaration *func_decl1 =
    func_call_exp->getAssociatedFunctionDeclaration();
  if(func_decl1) {

    func_decl1 = isSgFunctionDeclaration(func_decl1->get_definingDeclaration());

    if(func_decl1) {

      StringUtility::FileNameLibrary test
        = risc::tools::get_library(func_decl1->get_definition()->
                                       get_file_info()->get_file_id());

      if(test != "User") {

        return false;
      }
    }
  }

  return true;
}

std::vector<SgFunctionDeclaration*>
risc::tools::get_all_other_function_declarations(
    SgFunctionDeclaration *func_decl)
{
  std::vector<SgFunctionDeclaration*> func_decls;

  std::vector<SgNode*> temp = RoseNodes::get_func_decls();

  for(std::vector<SgFunctionDeclaration*>::iterator
      iter  = func_decls.begin();
      iter != func_decls.end();
      iter++) {

    SgFunctionDeclaration *func_temp = isSgFunctionDeclaration(*iter);

    if(func_decl->get_firstNondefiningDeclaration()
        == func_temp->get_firstNondefiningDeclaration()) {

      temp.push_back(func_temp);
    }
  }

  return func_decls;
}


SgFunctionDeclaration*
risc::tools::get_function_declaration_of_channel_call_dynamically(
    SgFunctionCallExp *func_call)
{
  // get the port variable in the in function call expression
  std::vector<SgNode*> var_refs
    = NodeQuery::querySubTree(func_call, V_SgVarRefExp);

  SgVarRefExp *port = NULL;
  for(std::vector<SgNode*>::iterator
      iter  = var_refs.begin();
      iter != var_refs.end();
      iter++) {

    SgVarRefExp *var_ref = isSgVarRefExp(*iter);

    if(tools::is_derived_from_sc_port_base(var_ref)) {

      port = var_ref;
      break;
    }
  }

  std::list<DynamicModule*> modules;
  SgClassType *class_type = isSgClassType(port->get_type());

  if(class_type) {

    SgFunctionDefinition *func_def
      = SageInterface::getEnclosingFunctionDefinition(func_call);

    SgFunctionDeclaration *func_decl = func_def->get_declaration();
    func_decl
      = isSgFunctionDeclaration(func_decl->get_firstNondefiningDeclaration());

    SgClassDefinition *module_name
      = SageInterface::getEnclosingClassDefinition(func_decl);

    get_all_components_of_type(
        ast_root,
        SageInterface::get_name(module_name->get_declaration()),
        modules);
  }

  DynamicPort *dyn_port = NULL;
  if(modules.empty() == false) {

    SgClassDeclaration *class_decl
      = isSgClassDeclaration(class_type->get_declaration());
    class_decl = isSgClassDeclaration(class_decl->get_definingDeclaration());

    for(std::list<DynamicModule*>::iterator
        iter_modules  = modules.begin();
        iter_modules != modules.end();
        iter_modules++) {

      DynamicModule *module = *iter_modules;

      for(std::list<DynamicPort*>::iterator
          iter_port  = module->ports_.begin();
          iter_port != module->ports_.end();
          iter_port++) {

        if((*iter_port)->name_ == port->unparseToString()) {

          dyn_port = *iter_port;
        }
      }
    }
  }

  DynamicModule *host = get_host_module_of_member(ast_root, dyn_port->address_);

  DynamicPrimChannel *channel = NULL;

  // Traversing the module hierarchy
  for(std::list<DynamicPrimChannel*>::iterator
      sub_channel_iter  = host->sub_channels_.begin();
      sub_channel_iter != host->sub_channels_.end();
      sub_channel_iter++) {

    if((*sub_channel_iter)->address_ == dyn_port->address_) {
      channel = (*sub_channel_iter);
    }
  }

  SgClassDefinition *channel_class_def = NULL;

  std::vector<SgNode*> class_defs = tools::RoseNodes::get_class_defs();

  for(std::vector<SgNode*>::iterator
      iter  = class_defs.begin();
      iter != class_defs.end();
      iter++) {

    SgClassDefinition *class_def = isSgClassDefinition(*iter);

    if(SageInterface::get_name(class_def->get_declaration())
          == channel->class_name_) {

      channel_class_def = class_def;
    }
  }

  std::vector<SgNode*> func_decls
    = NodeQuery::querySubTree(channel_class_def, V_SgFunctionDeclaration);

  SgFunctionDeclaration *asso_func_decl = NULL;

  SgBinaryOp *binary = isSgBinaryOp(func_call->get_function());

  if(binary) {

    SgMemberFunctionSymbol *func_symbol
      = isSgMemberFunctionRefExp(binary->get_rhs_operand())->get_symbol();
    asso_func_decl = func_symbol->get_declaration();

  } else {
    std::cout << "Error: No associated function declaration for: "
              << func_call->unparseToString() << std::endl;
    exit(1);
  }

  SgFunctionDeclaration *called_function = NULL;

  for(std::vector<SgNode*>::iterator
      iter  = func_decls.begin();
      iter != func_decls.end();
      iter++) {

    SgFunctionDeclaration *func_decl = isSgFunctionDeclaration(*iter);

    if(SageInterface::get_name(asso_func_decl)
          != SageInterface::get_name(*iter)) {
      continue;

    } else {

      called_function = func_decl;
      break;
    }
  }

  return called_function;
}


std::list<SgClassDefinition*>
risc::tools::get_channel_interfaces(SgClassDefinition *channel_def)
{
  std::list<SgClassDefinition*> interfaces;

  for(SgBaseClassPtrList::iterator
      base_iter  = channel_def->get_inheritances().begin();
      base_iter != channel_def->get_inheritances().end();
      base_iter++) {

    SgClassDeclaration *base_class_decl = (*base_iter)->get_base_class();
    base_class_decl
      = isSgClassDeclaration(base_class_decl->get_definingDeclaration());

    if(base_class_decl) {
      SgClassDefinition *base_class_def = base_class_decl->get_definition();

      if(base_class_def) {

        for(SgBaseClassPtrList::iterator
            base_base_iter  = base_class_def->get_inheritances().begin();
            base_base_iter != base_class_def->get_inheritances().end();
            base_base_iter++) {

          SgClassDeclaration *base_base_class_decl
            = (*base_base_iter)->get_base_class();

          base_base_class_decl
            = isSgClassDeclaration(
                  base_base_class_decl->get_definingDeclaration());

          if(base_base_class_decl) {
            SgClassDefinition *base_base_class_def
              = base_base_class_decl->get_definition();

            if(base_base_class_def
                == tools::SystemCDefinitions::get_sc_interface()) {

              interfaces.push_back(base_class_def);
            }
          }
        }
      }
    }
  }

  return interfaces;
}


std::list<SgClassDefinition*>
risc::tools::get_all_used_interfaces(bool only_in_user_domain)
{
  risc::Design *design = risc::get_design();

  std::list<SgClassDefinition*> interfaces;

#if 0
  #define ANALYZE_CHANNEL(TYPE, VARIABLE)                                                        \
  for(risc:: #TYPE ::iterator                                                                    \
      channel_iter  = design->#VARIABLE.begin();                                                 \
      channel_iter != design->#VARIABLE.end();                                                   \
      channel_iter++) {                                                                          \
    if(only_in_user_domain) {                                                                    \
      if((*channel_iter)->get_ast_node()->get_file_info()->get_file_id() == 0) {                 \
        continue;                                                                                \
      }                                                                                          \
    }                                                                                            \
    SgClassDefinition *channel_def = (*channel_iter)->get_ast_node();                            \
    std::list<SgClassDefinition*> channel_interfaces = get_channel_interfaces(channel_def);      \
    if(only_in_user_domain) {                                                                    \
      for(std::list<SgClassDefinition*>::iterator                                                \
          interface_iter  = channel_interfaces.begin();                                          \
          interface_iter != channel_interfaces.end();                                            \
          interface_iter++) {                                                                    \
        if((*interface_iter)->get_file_info()->get_file_id() == 0) {                             \
          interfaces.push_back(*interface_iter);                                                 \
        }                                                                                        \
      }                                                                                          \
    } else {                                                                                     \
      interfaces.insert(interfaces.end(), channel_interfaces.begin(), channel_interfaces.end()); \
    }                                                                                            \
  }

  ANALYZE_CHANNEL(PrimitiveChannelVector, primitive_channel_);
  ANALYZE_CHANNEL(HierarchicalChannelVector, hierarchical_channel_);

  #undef ANALYZE_CHANNEL


#endif

#if 1
  for(risc::PrimitiveChannelVector::iterator
      channel_iter  = design->primitive_channel_.begin();
      channel_iter != design->primitive_channel_.end();
      channel_iter++) {

    if(only_in_user_domain) {

      if((*channel_iter)->get_ast_node()->get_file_info()->get_file_id() == 0) {
        continue;
      }
    }

    SgClassDefinition *channel_def = (*channel_iter)->get_ast_node();
    std::list<SgClassDefinition*> channel_interfaces
      = get_channel_interfaces(channel_def);

    if(only_in_user_domain) {
      for(std::list<SgClassDefinition*>::iterator
          interface_iter  = channel_interfaces.begin();
          interface_iter != channel_interfaces.end();
          interface_iter++) {

        if((*interface_iter)->get_file_info()->get_file_id() == 0) {

          interfaces.push_back(*interface_iter);
        }
      }
    } else {

      interfaces.insert(
          interfaces.end(),
          channel_interfaces.begin(),
          channel_interfaces.end());
    }
  }

  for(risc::HierarchicalChannelVector::iterator
      channel_iter  = design->hierarchical_channel_.begin();
      channel_iter != design->hierarchical_channel_.end();
      channel_iter++) {

    if(only_in_user_domain) {

      if((*channel_iter)->get_ast_node()->get_file_info()->get_file_id() != 0) {
        continue;
      }
    }

    SgClassDefinition *channel_def = (*channel_iter)->get_ast_node();
    std::list<SgClassDefinition*> channel_interfaces
      = get_channel_interfaces(channel_def);

    if(only_in_user_domain) {

      for(std::list<SgClassDefinition*>::iterator
          interface_iter  = channel_interfaces.begin();
          interface_iter != channel_interfaces.end();
          interface_iter++) {

        if((*interface_iter)->get_file_info()->get_file_id() == 0) {

          interfaces.push_back(*interface_iter);
        }
      }
    } else {

      interfaces.insert(
          interfaces.end(),
          channel_interfaces.begin(),
          channel_interfaces.end());
    }
  }

#endif
  return interfaces;
}

std::list<SgFunctionDeclaration*>
risc::tools::get_interface_functions(
    SgClassDefinition *channel_def,
    bool only_annotated)
{
  std::list<SgFunctionDeclaration*> channel_interface_func_decls;

  std::list<SgClassDefinition*> interfaces
    = get_channel_interfaces(channel_def);

  std::vector<SgNode*> channel_func_decls
    = NodeQuery::querySubTree(channel_def, V_SgFunctionDeclaration);

  for(std::list<SgClassDefinition*>::iterator
      interface_iter  = interfaces.begin();
      interface_iter != interfaces.end();
      interface_iter++) {

    std::vector<SgNode*> interface_func_decls
      = NodeQuery::querySubTree(*interface_iter, V_SgFunctionDeclaration);

    for(std::vector<SgNode*>::iterator
        interface_func_decl_iter  = interface_func_decls.begin();
        interface_func_decl_iter != interface_func_decls.end();
        interface_func_decl_iter++) {

      SgFunctionDeclaration *interface_func_decl
        = isSgFunctionDeclaration(*interface_func_decl_iter);

      for(std::vector<SgNode*>::iterator
          channel_func_decl_iter  = channel_func_decls.begin();
          channel_func_decl_iter != channel_func_decls.end();
          channel_func_decl_iter++) {

        SgFunctionDeclaration *channel_func_decl
          = isSgFunctionDeclaration(*channel_func_decl_iter);

        if(SageInterface::get_name(interface_func_decl)
            == SageInterface::get_name(channel_func_decl)) {

          if(only_annotated) {

            if(is_annotated_function(channel_func_decl)) {
              channel_interface_func_decls.push_back(channel_func_decl);
            }
          } else {
            channel_interface_func_decls.push_back(channel_func_decl);
          }
        }
      }
    }
  }

  return channel_interface_func_decls;
}

std::list<SgFunctionCallExp*>
risc::tools::get_all_port_calls_to_annotated_functions(
    risc::sg::SegmentGraph &graph)
{
  std::list<SgFunctionCallExp*> func_decls;

  typedef boost::graph_traits<risc::sg::Graph>::vertex_iterator VertexIterator;
  std::pair<VertexIterator, VertexIterator> vp;
  for(vp = boost::vertices(graph.graph_);
      vp.first != vp.second;
      ++vp.first) {

    if(graph.graph_[*vp.first].inlined_segment_) {
      SgNode *boundary
        = const_cast<SgNode*>(graph.graph_[*vp.first].boundary_node_);

      SgFunctionCallExp *func_call_exp
        = isSgFunctionCallExp(boundary);

      if(func_call_exp) {
        func_decls.push_back(func_call_exp);
      }
    }
  }

  return func_decls;
}


std::list<SgFunctionDeclaration*>
risc::tools::get_all_implemented_interface_functions(
    bool only_in_user_domain,
    bool only_annotated)
{
  std::list<SgFunctionDeclaration*> channel_interface_functions;

  for(risc::PrimitiveChannelVector::iterator
      channel_iter  = risc::get_design()->primitive_channel_.begin();
      channel_iter != risc::get_design()->primitive_channel_.end();
      channel_iter++) {

    if(only_in_user_domain) {

      if((*channel_iter)->get_ast_node()->get_file_info()->get_file_id() != 0) {
        continue;
      }
    }
    std::list<SgFunctionDeclaration*> channel_interfaces
      = get_interface_functions(
            (*channel_iter)->get_ast_node(),
            only_annotated);

    channel_interface_functions.insert(
        channel_interface_functions.end(),
        channel_interfaces.begin(),
        channel_interfaces.end());
  }

  for(risc::HierarchicalChannelVector::iterator
      channel_iter  = risc::get_design()->hierarchical_channel_.begin();
      channel_iter != risc::get_design()->hierarchical_channel_.end();
      channel_iter++) {

    if(only_in_user_domain) {

      if((*channel_iter)->get_ast_node()->get_file_info()->get_file_id() != 0) {
        continue;
      }
    }

    std::list<SgFunctionDeclaration*> channel_interfaces
      = get_interface_functions(
            (*channel_iter)->get_ast_node(),
            only_annotated);

    channel_interface_functions.insert(
        channel_interface_functions.end(),
        channel_interfaces.begin(),
        channel_interfaces.end());
  }

  return channel_interface_functions;
}

bool risc::tools::is_annotated_function(SgFunctionDeclaration *func_decl)
{
  func_decl
    = isSgFunctionDeclaration(func_decl->get_firstNondefiningDeclaration());

  sg::FunctionAnnotationAttributes *function_annotation_attribute =
    dynamic_cast<sg::FunctionAnnotationAttributes*>(
        func_decl->getAttribute("function_annotation"));

  if(function_annotation_attribute) {
    return true;
  } else {
    return false;
  }
}

bool risc::tools::has_reference_parameter(SgFunctionDeclaration *func_decl)
{
  for(SgInitializedNamePtrList::iterator
      iter  = func_decl->get_args().begin();
      iter != func_decl->get_args().end();
      iter++) {

    if(isSgReferenceType((*iter)->get_type())) {
      return true;
    }
  }
  return false;
}

SgFunctionDeclaration*
risc::tools::get_first_non_defining_function_declaration(
    SgFunctionCallExp *func_call_exp)
{
  SgFunctionDeclaration *func_decl = NULL;
  if(tools::function_call_belongs_to_a_port(func_call_exp)) {
    if(risc::get_design()->perform_static_analysis()) {

      func_decl = tools::get_function_declaration_of_channel_call_statically(
                             func_call_exp);
    } else {

      func_decl = tools::get_function_declaration_of_channel_call_dynamically(
                             func_call_exp);
    }

    func_decl = isSgFunctionDeclaration(
                    func_decl->get_firstNondefiningDeclaration());

  } else {

    func_decl = func_call_exp->getAssociatedFunctionDeclaration();
    if(func_decl) {

      func_decl = isSgFunctionDeclaration(
                      func_decl->get_firstNondefiningDeclaration());
    }
  }
  return func_decl;
}

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
