#include <iostream>

#include <libgen.h>
#include <cstring>

#include "risc.h"

extern DynamicModule *ast_root;

int main(int argc, char* argv[])
{
  // Parse library-level command-line options
  risc::tools::ParseStatus status;
  risc::tools::ParseCommandlineOptions(argc, argv, &status);

  std::cout << std::endl << "start processing file ..." << std::endl;

  risc::Design d(argc,argv);

  if(ast_root) {
    ast_root->print_tree(0);
  } else {
    std::cout << "No root defined" << std::endl;
  }

  risc::sg::SegmentGraph::FunctionCallBoundarySet function_call_boundary(
      risc::tools::SystemCDefinitions::get_sc_wait_funcs());
  risc::sg::SegmentGraph sg(function_call_boundary);

  for(risc::HierarchicalChannelVectorIter
      iter_channel  = d.hierarchical_channel_.begin();
      iter_channel != d.hierarchical_channel_.end();
      iter_channel++) {

    std::cout << "Channel: " <<  (*iter_channel)->get_name() << std::endl;

    for(risc::ThreadVectorIter
        iter_thread  = (*iter_channel)->threads_.begin();
        iter_thread != (*iter_channel)->threads_.end();
        iter_thread++) {

      std::cout << "  " << (*iter_thread)->get_name() << std::endl;
      sg.add_function(*iter_thread);
    }

    std::cout << std::endl;
  }

  for(risc::ModuleVectorIter
      iter_module  = d.module_definitions_.begin();
      iter_module != d.module_definitions_.end();
      iter_module++) {

    std::cout << "Module: " <<  (*iter_module)->get_name() << std::endl;

    for(risc::ThreadVectorIter
        iter_thread  = (*iter_module)->threads_.begin();
        iter_thread != (*iter_module)->threads_.end();
        iter_thread++) {

      std::cout << "  " << (*iter_thread)->get_name() << std::endl;
      sg.add_function(*iter_thread);
    }

    std::cout << std::endl;
  }

  // getting the name of the file without extension
  SgSourceFile *file
    = isSgSourceFile(&risc::get_project()->get_file(0));
  char filename[strlen(file->getFileName().c_str()) + 1];

  strcpy(filename, file->getFileName().c_str());

  std::string filename_without_extension(basename(filename));

  // deleting the appending .cpp
  filename_without_extension.erase(filename_without_extension.length()-4, 4);

  risc::sg::DataConflictTable data(sg, d.path_instance_mapper_, true);
  if(d.perform_static_analysis()) {
    sg.print_graph("segment_graph_" + filename_without_extension + "_static.dot");
    sg.print_graph_read_write_access("segment_graph_read_write_"
        + filename_without_extension + "_static.dot");
    data.print_conflict_table("data_conflict_table_"
        + filename_without_extension + "_static.html");
  } else {
    sg.print_graph("segment_graph_" + filename_without_extension + "_dynamic.dot");
    sg.print_graph_read_write_access("segment_graph_read_write_"
        + filename_without_extension + "_dynamic.dot");
    data.print_conflict_table("data_conflict_table_"
        + filename_without_extension + "_dynamic.html");
  }

  risc::sg::EventConflictTable event(sg, d.path_instance_mapper_, true);
  if(d.perform_static_analysis()) {
    event.print_conflict_table("event_conflict_table_"
        + filename_without_extension + "_static.html");
  } else {

    event.print_conflict_table("event_conflict_table_"
        + filename_without_extension + "_dynamic.html");
  }

  risc::sg::TimeAdvanceTable time_advance_table(sg);

  std::cout << "start instrumenting" << std::endl;
  risc::inst::instrumentor(d, sg, &data, &event, &time_advance_table, &status);
  std::cout << "end instrumenting" << std::endl;

  std::cout << std::endl << "end processing file ..." << std::endl;

  return 0;
}

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
